void initrandom(void);
void seedrandom(unsigned char *, unsigned int);
unsigned char randombyte(void);
void closerandom(void);
unsigned char *getstring(int);
void sn512string(unsigned char *, int, unsigned char *);
